/* -*- Mode: C; c-basic-offset:4 ; -*- */
/*
 *  (C) 2001 by Argonne National Laboratory.
 *      See COPYRIGHT in top-level directory.
 */

#if !defined(MPICH_MPIDI_CH3_IMPL_H_INCLUDED)
#define MPICH_MPIDI_CH3_IMPL_H_INCLUDED

#ifdef USE_WINCONF_H
#include "winmpidi_ch3i_rdma_conf.h"
#else
#include "mpidi_ch3i_rdma_conf.h"
#endif
#include "mpidimpl.h"

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_ASSERT_H
#include <assert.h>
#endif

#define MPIDI_CH3I_SPIN_COUNT_DEFAULT   100
#define MPIDI_CH3I_YIELD_COUNT_DEFAULT  5000

#define MPIDI_CH3I_READ_STATE_IDLE    0
#define MPIDI_CH3I_READ_STATE_READING 1


typedef struct MPIDI_CH3I_Process_s
{
    MPIDI_CH3I_Process_group_t * pg;
    MPIDI_VC *vc;
}
MPIDI_CH3I_Process_t;

extern MPIDI_CH3I_Process_t MPIDI_CH3I_Process;

#define MPIDI_CH3I_SendQ_enqueue(vc, req)				\
{									\
    /* MT - not thread safe! */						\
    MPIDI_DBG_PRINTF((50, FCNAME, "SendQ_enqueue vc=%08p req=0x%08x",	\
	              vc, req->handle));		\
    req->ch3.next = NULL;						\
    if (vc->rdma.sendq_tail != NULL)					\
    {									\
	vc->rdma.sendq_tail->ch3.next = req;				\
    }									\
    else								\
    {									\
	vc->rdma.sendq_head = req;					\
    }									\
    vc->rdma.sendq_tail = req;						\
}

#define MPIDI_CH3I_SendQ_enqueue_head(vc, req)				     \
{									     \
    /* MT - not thread safe! */						     \
    MPIDI_DBG_PRINTF((50, FCNAME, "SendQ_enqueue_head vc=%08p req=0x%08x", \
	              vc, req->handle));		     \
    req->ch3.next = vc->rdma.sendq_tail;					     \
    if (vc->rdma.sendq_tail == NULL)					     \
    {									     \
	vc->rdma.sendq_tail = req;					     \
    }									     \
    vc->rdma.sendq_head = req;						     \
}

#define MPIDI_CH3I_SendQ_dequeue(vc)					\
{									\
    /* MT - not thread safe! */						\
    MPIDI_DBG_PRINTF((50, FCNAME, "SendQ_dequeue vc=%08p req=0x%08x",	\
	              vc, vc->rdma.sendq_head));		\
    vc->rdma.sendq_head = vc->rdma.sendq_head->ch3.next;			\
    if (vc->rdma.sendq_head == NULL)					\
    {									\
	vc->rdma.sendq_tail = NULL;					\
    }									\
}

#define MPIDI_CH3I_SendQ_head(vc) (vc->rdma.sendq_head)

#define MPIDI_CH3I_SendQ_empty(vc) (vc->rdma.sendq_head == NULL)

/* RDMA channel interface */

int MPIDI_CH3I_Progress_init(void);
int MPIDI_CH3I_Progress_finalize(void);
int MPIDI_CH3I_Request_adjust_iov(MPID_Request *, MPIDI_msg_sz_t);

int MPIDI_CH3I_read_progress(int blocking, MPIDI_VC **vc_pptr, int *num_bytes_ptr);
int MPIDI_CH3I_write_progress();
int MPIDI_CH3I_post_read(MPIDI_VC *vc, void *buf, int len);
int MPIDI_CH3I_post_readv(MPIDI_VC *vc, MPID_IOV *iov, int n);

/* RDMA implementation interface */

/* init must allocate RDMA memory and initialize the queues and other structures inside it */
int MPIDI_CH3I_RDMA_init_process_group(int * has_parent);
int MPIDI_CH3I_RMDA_init();
/* finalize releases the RDMA memory and any other cleanup */
int MPIDI_CH3I_RMDA_finalize();

/* put_datav attempts to put the data described by the iov in the RDMA memory of the remote process
   specified by the vc */
int MPIDI_CH3I_RDMA_put_datav(MPIDI_VC *vc, MPID_IOV *iov, int n, int *num_bytes_ptr);
/* set_read_ready signals the remote process that a put has completed. */
/*int MPIDI_CH3I_RDMA_set_read_ready(MPIDI_VC *vc);*/
/* read_ready returns true when data can be read locally.  set_read_ready on one process will cause
   read_ready on another process to succeed. */
/*int MPIDI_CH3I_RDMA_read_ready(MPIDI_VC *vc);*/
/* read_datav reads data from the local RDMA memory and copies it to the user buffer described by the iov. */
int MPIDI_CH3I_RDMA_read_datav(MPIDI_VC *vc, MPID_IOV *iov, int n, int *num_bytes_ptr);

#endif /* !defined(MPICH_MPIDI_CH3_IMPL_H_INCLUDED) */
